searchData={"items":[{"type":"module","title":"Pacer.Config","doc":"The Pacer.Config module provides functions for\nextracting user-provided configuration values specific\nto Pacer.","ref":"Pacer.Config.html"},{"type":"function","title":"Pacer.Config.batch_telemetry_options/1","doc":"Fetches configuration options for extending the metadata provided in the\n`[:pacer, :execute_vertex, :start | :stop | :exception]` events for batched resolvers.\n\nThe configuration must be set under the key `:batch_telemetry_options` at the application\nlevel (i.e., `Application.get_env(:pacer, :batch_telemetry_options)`) or when defining the\nworkflow itself (`use Pacer.Workflow, batch_telemetry_options:  `).\n\nThe batch_telemetry_options defined by the user must be either:\n  - a keyword list, or\n  - a {module, function, args} mfa tuple; when invoked, this function must return a keyword list\n\nThe keyword list of values returned by the mfa-style config, or the hardcoded keyword list, is\nfetched and converted into a map that gets merged into the telemetry event metadata for batched resolvers.","ref":"Pacer.Config.html#batch_telemetry_options/1"},{"type":"function","title":"Pacer.Config.fetch_batch_telemetry_options/1","doc":"Takes the batch_telemetry_options configuration, invoking mfa-style config if available,\nand converts the batch_telemetry_options keyword list into a map that gets merged into\nthe metadata for the `[:pacer, :execute_vertex, :start | :stop | :exception]` events for\nbatched resolvers.","ref":"Pacer.Config.html#fetch_batch_telemetry_options/1"},{"type":"module","title":"Pacer.Docs","doc":"This module is responsible for extracting information provided to fields\nin a Pacer.Workflow `graph` definition and turning that information into\ndocs that can be rendered into a module's documentation.","ref":"Pacer.Docs.html"},{"type":"macro","title":"Pacer.Docs.generate/0","doc":"","ref":"Pacer.Docs.html#generate/0"},{"type":"function","title":"Pacer.Docs.write_default/1","doc":"","ref":"Pacer.Docs.html#write_default/1"},{"type":"function","title":"Pacer.Docs.write_dependencies/1","doc":"","ref":"Pacer.Docs.html#write_dependencies/1"},{"type":"function","title":"Pacer.Docs.write_doc/1","doc":"","ref":"Pacer.Docs.html#write_doc/1"},{"type":"module","title":"Pacer.Workflow","doc":"Dependency Graph-Based Workflows With Robust Compile Time Safety & Guarantees","ref":"Pacer.Workflow.html"},{"type":"module","title":"Motivations - Pacer.Workflow","doc":"`Pacer.Workflow` is designed for complex workflows where many interdependent data points need to be\nstitched together to provide a final result, specifically workflows where each data point needs to be\nloaded and/or calculated using discrete, application-specific logic.\n\nTo create a struct backed by Pacer.Workflow, invoke `use Pacer.Workflow` at the top of your module and use\nthe `graph/1` macro, which is explained in more detail in the docs below.\n\nNote that when using `Pacer.Workflow`, you can pass the following options:\n* `:generate_docs?` (`t:boolean/0`) - By invoking `use Pacer.Workflow`, Pacer will automatically generate module documentation for you. It will create a section\n  titled `Pacer Fields` in your moduledoc, either by creating a moduledoc for you dynamically or appending this section to\n  any existing module documentation you have already provided.\n\n  To opt out of this feature, when you `use Pacer.Workflow`, set this option to false.\n\n  The default value is `true`.\n\n\n\nThe following is a list of the main ideas and themes underlying `Pacer.Workflow`\n\n##","ref":"Pacer.Workflow.html#module-motivations"},{"type":"module","title":"1. `Workflow`s Are Dependency Graphs - Pacer.Workflow","doc":"`Pacer.Workflow`s are backed by dependency graphs (specifically represented as directed acyclic graphs) that are constructed at compile-time.\nYour `Workflow`s will define a set of data points, represented as `field`s (see below); each `field` must explicitly define\nthe dependencies it has on other fields in the `Workflow`. For example, if we have a workflow where we load\na set of users and then fire off some requests to a 3rd party service to fetch some advertisements for those users,\nour `Workflow` might look something like this:\n\n```elixir\ndefmodule UserAdsWorkflow do\n  use Pacer.Workflow\n\n  graph do\n    field(:users)\n    field(:user_advertisements, resolver: &Ads.fetch_user_ads/1, dependencies: [:users])\n  end\nend\n```\n\nWhy is the dependency graph idea important here?\n\nIn the above, simplified example with only two fields, there may not be a need to define a dependency graph\nbecause we can look at the two fields and immediately realize that we first need to have the set of `users`\nbefore we can make the call to load `:user_advertisements`.\n\nHowever, in complex workflows with dozens or even hundreds of data points, if we were to try to manage what data points\nneed to be loaded in which order manually, it would be a daunting and time-consuming task. Not only that, we would also\nrun the risk of getting the ordering wrong, AND/OR when new data points are added or removed in the future, that we would\nneed to manually rearrange things for each data point to be loaded in the correct order.\n\nThe result is untenable for workflows of sufficient size.\n\nThis is where dependency graphs come in to play. By forcing you to explicitly declare the other fields that\nyou depend on in the workflow, `Pacer.Workflow` can build out a dependency graph and figure out how to schedule the\nexecution of each of your resolver functions (see below for more details on resolvers) so that each function\nwill only be called when its dependencies are ready. That eliminates the need to manually rearrange calls in\nyour codebase, and also allows you to have discrete, single-purpose resolvers that can be rigorously unit-tested\nagainst a known, constrained set of inputs.\n\n##","ref":"Pacer.Workflow.html#module-1-workflow-s-are-dependency-graphs"},{"type":"module","title":"2. Batched, Parallel Requests to Disparate External Systems (3rd-party APIs, Database Calls, etc.) - Pacer.Workflow","doc":"`Pacer.Workflow`s also allow users to fire off potentially high-latency calls in parallel to reduce overall\nlatency of a `Workflow`. To do so, we can use the `batch/2` macro inside of your `graph` definition. One caveat\nto this, however, is that _fields inside of a batch definition must not have any dependencies on other fields\ninside the same batch_.\n\nBatches are nice to use when a workflow has multiple high-latency requests that need to be made. Batching the\nrequests together, when possible, will fire off the requests in parallel. The requests can be to disparate,\nunrelated services, APIs, and external systems including databases and/or caches.\n\nNote: `batch`es should not be confused with `batch loading` data in the sense that, for example, GraphQL batches\nare used where users may provide a set of ids, etc., for related entities and the batch processing loads all of or\nas many of those entities in a single request rather than making a single request per entity. `Pacer.Workflow` batches\ncan be used to do so in roughly the same way, but that choice is left up to the user and the implementation.\nThe key idea of a `batch` here is that you have multiple (potentially) high-latency requests that you want to execute\ntogether (in parallel), rather than saying \"I have a set of entities that I want to load as a batch request\".\n\nFor example, if we go back to the earlier example of a user-based workflow where we load a set of users and fetch\nadvertisements for those users, if we add in another request to, say, an analytics service to get some more data on\nthe set of users we have just loaded, we can do that in a batch as follows:\n\n```elixir\ndefmodule UserAdsWorkflow do\n  use Pacer.Workflow\n\n  graph do\n    field(:users)\n    batch :requests do\n      field(:user_advertisements, resolver: &Ads.fetch_user_ads/1, dependencies: [:users])\n      field(:analytics, resolver: &Analytics.analyze_users/1, dependencies: [:users])\n    end\n  end\nend\n```\n\nNow, rather than those two requests being fired sequentially (and thereby boosting the latency of the workflow to be\nequal to the latency of the ads request _plus_ the latency of the analytics request,\nthe latency will instead be capped at the slowest of the two requests).\n\n##","ref":"Pacer.Workflow.html#module-2-batched-parallel-requests-to-disparate-external-systems-3rd-party-apis-database-calls-etc"},{"type":"module","title":"3. Compile-Time Safety And Guarantees - Pacer.Workflow","doc":"The third motivating factor behind `Pacer.Workflow` is to provide a robust set of compile-time safety mechanisms.\nThese include:\n  - Detecting and preventing cyclical dependencies in the dependency graph defined by the workflow\n  - Preventing \"reflexive\" dependencies, where a field depends on itself\n  - Detecting invalid options on `field`s and `batch`es\n  - Preventing a single module from defining more than one `Workflow`\n  - Detecting duplicate field definitions in a graph\n  - Ensuring that resolver definitions fit the contract required by `Pacer.Workflow` (a 1-arity function that takes a map)\n  - Detecting dependencies on fields that do not exist in the graph definition\n  - Requiring fields defined inside of a batch to have a resolver function defined\n\n`Pacer.Workflow` strives to provide helpful error messages to the user at compile time when it detects any issues\nand tries to direct the user on what went wrong, why, and how to fix the issue.\n\nThe compile-time safety can prevent a whole class of issues at runtime, and also allows the dependency graph\nto be computed once at compile time. Building the dependency graph at compile time allows Pacer to cache the\nresults of the graph and make those results accessible at runtime so your application does not have to incur\nthe cost of building out the dependency graph at runtime.","ref":"Pacer.Workflow.html#module-3-compile-time-safety-and-guarantees"},{"type":"module","title":"Summary - Pacer.Workflow","doc":"`Pacer.Workflow` provides the ability to explicitly declare a dependency graph, where the nodes in the\ngraph map to fields in a struct defined via the `graph/1` API.\n\nThe key idea behind `Pacer.Workflow` is that it enables users to create Elixir structs that serve as containers\nof loosely-related fields, where the fields in the struct have dependencies between other fields in the struct.\n\nA \"dependency\", in the sense it is used here, means that one field relies on another field's value being readily\navailable and loaded in memory before its own value can be computed or loaded. For example, if you have a struct\n`%MyStruct{field_a: 1, field_b:  }`, `:field_b` is dependent on `:field_a`'s value already\nbeing present before it can be calculated.\n\nThe example given above can be solved in a more straightforward way, by having a simple function to build out the\nentire struct given `:field_a`'s value as input, i.e.:\n\n```elixir\ndef build_my_struct(field_a_input) do\n  %MyStruct{field_a: field_a_input, field_b: field_a_input + 1}\nend\n```\n\nWhile conceptually simple, this pattern becomes more difficult to maintain when additional fields are added with dependencies between each other.\n\n`Pacer.Workflow` addresses this problem by forcing users to explicitly declare the dependencies between fields up front, at compile time.\nOnce the fields and dependencies have been declared, `Pacer.Workflow` can build a dependency graph, which allows the graph to solve for\nthe problem of dependency resolution by answering the question: Which fields need to be available when and in what order do they need to be executed?\n\nThere are a few key concepts to know in order to build out a `Pacer.Workflow`-backed struct:","ref":"Pacer.Workflow.html#module-summary"},{"type":"module","title":"Fields - Pacer.Workflow","doc":"A field can be defined within a graph definition with the `field/2` macro. A field\nmaps one-to-one to keys on the struct generated by the graph definition. Fields are\nhow you explicitly declare the dependencies each field has on other fields within the\nsame graph. You do this by providing a list of dependencies as atoms to the `field/2`\nmacro:\n\n  ```elixir\n  graph do\n    field(:field_one)\n    field(:field_two)\n    field(:my_dependent_field, resolver: &MyResolver.resolve/1 dependencies: [:field_one, :field_two])\n  end\n  ```\n\nIf the `:dependencies` option is not given, it defaults to an empty list and effectively means\nthat the field has no dependencies. This may be the case when the value for the field meets one\nof the following conditions:\n\n  - The value is a constant\n  - The value is already available and accessible in memory when creating the struct\n\nFields that do explicitly declare at least one dependency MUST also pass in a `:resolver` option.\nSee the [Resolvers](#resolvers) section below for more details.\n\nAdditionally, fields may declare a default value by passing a default to the `:default` option key:\n\n  ```elixir\n  graph do\n    field(:my_field, default: 42)\n  end\n  ```","ref":"Pacer.Workflow.html#module-fields"},{"type":"module","title":"Resolvers - Pacer.Workflow","doc":"Resolvers are 1-arity functions that take in the values from dependencies as input and return\nthe value that should be placed on the struct key for the associated `field`. Resolvers are\nfunction definitions that `Pacer.Workflow` can use to incrementally compute all values needed.\n\nFor example, for a graph definition that looks like this:\n\n  ```elixir\n  defmodule MyGraph do\n    use Pacer.Workflow\n\n    graph do\n      field(:field_one)\n      field(:dependent_field, resolver: &__MODULE__.resolve/1, dependencies: [:field_one])\n    end\n\n    def resolve(inputs) do\n      IO.inspect(inputs.field_one_value, label: \"Received field_one's value\")\n    end\n  end\n  ```\n\nResolver functions will always be called with a map that contains the values for fields declared as dependencies.\nIn the above example, that means if we have a struct `%MyGraph{field_one: 42}`, the resolver will be invoked with\n`%{field_one: 42}`.\n\nKeep in mind that if you declare any dependencies, you MUST also declare a resolver.","ref":"Pacer.Workflow.html#module-resolvers"},{"type":"module","title":"Batches - Pacer.Workflow","doc":"Batches can be defined using the `batch/3` macro.\n\nBatches allow users to group together a set of fields whose resolvers can and should be run in\nparallel. The main use-case for batches is to reduce running time for fields whose resolvers can\nhave high-latencies. This generally means that batches are useful to group together calls that hit\nthe network in some way.\n\nBatches do impose some more restrictive constraints on users, however. For example, all fields\ndefined within a batch MUST NOT declare dependencies on any other field in the same batch. This\nis because the resolvers will run concurrently with one another, so there is no way to guarantee\nthat a field within the same batch will have a value ready to use and pass to a separate resolver\nin the same batch. In scenarios where you find this happening, `Pacer.Workflow` will raise a compile time\nerror and you will need to rearrange your batches, possibly creating two separate batches or forcing\none field in the batch to run sequentially as a regular field outside of a `batch` block.\n\nBatches must also declare a name and fields within a batch must define a resolver.\nBatch names must also be unique within a single graph definition. Resolvers are required\nfor fields within a batch regardless of whether or not the field has any dependencies.\n\nEx.:\n\n  ```elixir\n  defmodule MyGraphWithBatches do\n    use Pacer.Workflow\n\n    graph do\n      field(:regular_field)\n\n      batch :http_requests do\n        field(:request_one, resolver: &__MODULE__.resolve/1)\n        field(:request_two, resolver: &__MODULE__.resolve/1, dependencies: [:regular_field])\n      end\n\n      field(:another_field, resolver: &__MODULE__.simple_resolver/1, dependencies: [:request_two])\n    end\n\n    def resolve(_) do\n      IO.puts(\"Simulating HTTP request\")\n    end\n\n    def simple_resolver(_), do: :ok\n  end\n  ```\n\nNotes:\n\nThe order fields are defined in within a `graph` definition does not matter. For example, if you have a field `:request_one` that depends\non another field `:request_two`, the fields can be declared in any order.","ref":"Pacer.Workflow.html#module-batches"},{"type":"module","title":"Telemetry - Pacer.Workflow","doc":"Pacer provides two levels of granularity for workflow telemetry: one at the entire workflow level, and one at the resolver level.\n\nFor workflow execution, Pacer will trigger the following telemetry events:\n\n- `[:pacer, :workflow, :start]`\n    - Measurements include: `%{system_time: integer(), monotonic_time: integer()}`\n    - Metadata provided: `%{telemetry_span_context: term(), workflow: module()}`, where the `workflow` key contains the module name for the workflow being executed\n- `[:pacer, :workflow, :stop]`\n   - Measurements include: `%{duration: integer(), monotonic_time: integer()}`\n   - Metadata provided: `%{telemetry_span_context: term(), workflow: module()}`, where the `workflow` key contains the module name for the workflow being executed\n- `[:pacer, :workflow, :exception]`\n    - Measurements include: `%{duration: integer(), monotonic_time: integer()}`\n    - Metadata provided: %{kind: :throw | :error | :exit, reason: term(), stacktrace: list(), telemetry_span_context: term(), workflow: module()}, where the `workflow` key contains the module name for the workflow being executed\n\nAt the resolver level, Pacer will trigger the following telemetry events:\n\n- `[:pacer, :execute_vertex, :start]`\n    - Measurements and metadata similar to `:workflow` start event, with the addition of the `%{field: atom()}` value passed in metadata. The `field` is the name of the field for which the resolver is being executed.\n- `[:pacer, :execute_vertex, :stop]`\n    - Measurements and metadata similar to `:workflow` stop event, with the addition of the `%{field: atom()}` value passed in metadata. The `field` is the name of the field for which the resolver is being executed.\n- `[:pacer, :execute_vertex, :exception]`\n    - Measurements and metadata similar to `:workflow` exception event, with the addition of the `%{field: atom()}` value passed in metadata. The `field` is the name of the field for which the resolver is being executed.\n\nAdditionally, for `[:pacer, :execute_vertex]` events fired on batched resolvers (which will run in parallel processes), users can provide their own metadata through configuration.\n\nUsers may provide either a keyword list of options which will be merged into the `:execute_vertex` event metadata, or an MFA `{mod, fun, args}` tuple that points to a function which\nreturns a keyword list that will be merged into the `:execute_vertex` event metadata.\n\nThere are two routes for configuring these telemetry options for batched resolvers: in the application environment using the `:pacer, :batch_telemetry_options` config key, or\non the individual workflow modules themselves by passing `:batch_telemetry_options` when invoking `use Pacer.Workflow`.\nConfiguration defined at the workflow module will override configuration defined in the application environment.\n\nHere are a couple of examples:\n\n#","ref":"Pacer.Workflow.html#module-telemetry"},{"type":"module","title":"User-Provided Telemetry Metadata for Batched Resolvers in Applicaton Config - Pacer.Workflow","doc":"```elixir\n# In config.exs (or whatever env config file you want to target):\n\nconfig :pacer, :batch_telemetry_options, application_name: MyApp","ref":"Pacer.Workflow.html#module-user-provided-telemetry-metadata-for-batched-resolvers-in-applicaton-config"},{"type":"module","title":"When you invoke a workflow with batched resolvers now, you will get `%{application_name: MyApp}` merged into your - Pacer.Workflow","doc":"","ref":"Pacer.Workflow.html#module-when-you-invoke-a-workflow-with-batched-resolvers-now-you-will-get-application_name-myapp-merged-into-your"},{"type":"module","title":"event metadata in the `[:pacer, :execute_vertex, :start | :stop | :exception]` events. - Pacer.Workflow","doc":"```\n\n#","ref":"Pacer.Workflow.html#module-event-metadata-in-the-pacer-execute_vertex-start-stop-exception-events"},{"type":"module","title":"User-Provided Telemetry Metadata for Batched Resolvers at the Workflow Level - Pacer.Workflow","doc":"```elixir\ndefmodule MyWorkflow do\n  use Pacer.Workflow, batch_telemetry_options: [extra_context: \"some context from my application\"]\n\n  graph do\n    field(:a)\n\n    batch :long_running_requests do\n      field(:b, dependencies: [:a], resolver: &Requests.trigger_b/1, default: nil)\n      field(:c, dependencies: [:a], resolver: &Requests.trigger_c/1, default: nil)\n    end\n  end\nend","ref":"Pacer.Workflow.html#module-user-provided-telemetry-metadata-for-batched-resolvers-at-the-workflow-level"},{"type":"module","title":"Now when you invoke `Pacer.execute(MyWorkflow)`, you will get `%{extra_context: \"some context from my application\"}` - Pacer.Workflow","doc":"","ref":"Pacer.Workflow.html#module-now-when-you-invoke-pacer-execute-myworkflow-you-will-get-extra_context-some-context-from-my-application"},{"type":"module","title":"merged into the metadata for the `[:pacer, :execute_vertex, :start | :stop | :exception]` events for fields `:b` and `:c` - Pacer.Workflow","doc":"```\n\nNote that you can also provide an MFA tuple that points to a module/function that returns a keyword list of options to be\ninjected into the metadata on `:execute_vertex` telemetry events for batched resolvers. This allows users to execute code at runtime\nto inject dynamic values into the metadata. Users may use this to inject things like span_context from the top-level workflow process\ninto the parallel processes that run the batch resolvers. This lets you propagate context from, i.e., a process dictionary at the top-level\ninto the sub-processes:\n\n```elixir\ndefmodule MyApp.BatchOptions do\n  def inject_context do\n    [span_context: MyTracingLibrary.Tracer.current_context()]\n  end\nend","ref":"Pacer.Workflow.html#module-merged-into-the-metadata-for-the-pacer-execute_vertex-start-stop-exception-events-for-fields-b-and-c"},{"type":"module","title":"Use this function to inject span context by configuring it at the workflow level or in the application environment - Pacer.Workflow","doc":"","ref":"Pacer.Workflow.html#module-use-this-function-to-inject-span-context-by-configuring-it-at-the-workflow-level-or-in-the-application-environment"},{"type":"module","title":"In config.exs: - Pacer.Workflow","doc":"config :pacer, :batch_telemetry_options, {MyApp.BatchOptions, :inject_context, []}\n```","ref":"Pacer.Workflow.html#module-in-config-exs"},{"type":"macro","title":"Pacer.Workflow.batch/3","doc":"The batch/3 macro is to be invoked when grouping fields with resolvers that will run in parallel.\n\nReminder:\n  - The batch must be named and unique.\n  - The fields within the batch must not have dependencies on one another since they will run concurrently.\n  - The fields within the batch must each declare a resolver function.\n\n__NOTE__: In general, only batch fields whose resolvers contain potentially high-latency operations, such as network calls.\n\nExample\n```elixir\n  defmodule MyValidGraph do\n    use Pacer.Workflow\n\n    graph do\n      field(:custom_field)\n\n      batch :http_requests do\n        field(:request_1, resolver: &__MODULE__.do_work/1, dependencies: [:custom_field])\n        field(:request_2, resolver: &__MODULE__.do_work/1, dependencies: [:custom_field])\n        field(:request_3, resolver: &__MODULE__.do_work/1)\n      end\n    end\n\n    def do_work(_), do: :ok\n  end\n```\n\nField options for fields defined within a batch have one minor requirement difference\nfrom fields not defined within a batch: batched fields MUST always define a resolver function,\nregardless of whether or not they define any dependencies.","ref":"Pacer.Workflow.html#batch/3"},{"type":"macro","title":"Batch Field Options - Pacer.Workflow.batch/3","doc":"* `:guard` (function of arity 1) - A guard is a 1-arity function that takes in a map with the field's dependencies and returns either true or false.\n  If the function returns `false`, it means there is no work to do and thus no reason to spin up another process\n  to run the resolver function. In this case, the field's default value is returned.\n  If the function returns `true`, the field's resolver will run in a separate process.\n\n* `:dependencies` (list of `t:atom/0`) - A list of dependencies from the graph.\n  Dependencies are specified as atoms, and each dependency must\n  be another field in the same graph.\n\n  Remember that cyclical dependencies are strictly not allowed,\n  so fields cannot declare dependencies on themselves nor on\n  any other field that has already declared a dependency on the\n  current field.\n\n  If the `dependencies` option is not given, it defaults to an\n  empty list, indicating that the field has no dependencies.\n  This will be the case if the field is a constant or can be\n  constructed from values already available in the environment.\n\n  The default value is `[]`.\n\n* `:doc` (`t:String.t/0`) - Allows users to document the field and provide background and/or context\n  on what the field is intended to be used for, what kind of data the field\n  contains, and how the data for the field is constructed.\n\n* `:resolver` (function of arity 1) - Required. A resolver is a 1-arity function that specifies how to calculate\n  the value for a field.\n\n  The argument passed to the function will be a map that contains\n  all of the field's declared dependencies.\n\n  For example, if we have a field like this:\n\n  ```elixir\n  field(:request, resolver: &RequestHandler.resolve/1, dependencies: [:api_key, :url])\n  ```\n\n  The resolver `RequestHandler.resolve/1` would be passed a map that looks like this:\n  ```elixir\n  %{api_key: \" \", url: \"https://some.endpoint.com\"}\n  ```\n\n  If the field has no dependencies, the resolver will receive an empty map. Note though\n  that resolvers are only required for fields with no dependencies if the field is inside\n  of a batch. If your field has no dependencies and is not inside a batch, you can skip\n  defining a resolver and initialize your graph struct with a value that is either constant\n  or readily available where you are constructing the struct.\n\n  The result of the resolver will be placed on the graph struct under the field's key.\n\n  For the above, assuming a graph that looks like this:\n\n  ```elixir\n  defmodule MyGraph do\n    use Pacer.Workflow\n\n    graph do\n      field(:api_key)\n      field(:url)\n      field(:request, resolver: &RequestHandler.resolve/1, dependencies: [:api_key, :url])\n    end\n  end\n  ```\n\n  Then when the `RequestHandler.resolve/1` runs an returns a value of, let's say, `%{response: \"important response\"}`,\n  your graph struct would look like this:\n\n  ```elixir\n  %MyGraph{\n    api_key: \" \",\n    url: \"https://some.endpoint.com\",\n    request: %{response: \"important response\"}\n  }\n  ```\n\n* `:default` (`t:term/0`) - Required. The default value for the field. If no default is given, the default value becomes\n  `#Pacer.Workflow.FieldNotSet `.\n\n* `:virtual?` (`t:boolean/0`) - A virtual field is used for intermediate or transient computation steps during the workflow and becomes a\n  node in the workflow's graph, but does not get returned in the results of the workflow execution.\n\n  In other words, virtual keys will not be included in the map returned by calling `Pacer.Workflow.execute/1`.\n\n  The intent of a virtual field is to allow a spot for intermediate and/or transient calculation steps but\n  to avoid the extra memory overhead that would be associated with carrying these values downstream if, for example,\n  the map returned from `Pacer.Workflow.execute/1` is stored in a long-lived process state; intermediate or transient\n  values can cause unnecessary memory bloat if they are carried into process state where they are not neeeded.\n\n  The default value is `false`.","ref":"Pacer.Workflow.html#batch/3-batch-field-options"},{"type":"macro","title":"Batch options - Pacer.Workflow.batch/3","doc":"* `:on_timeout` (`t:atom/0`) - Required. The task that is timed out is killed and returns {:exit, :timeout}.\n  This :kill_task option only exits the task process that fails and not the process that spawned the task. The default value is `:kill_task`.\n\n* `:timeout` (`t:non_neg_integer/0`) - Required. The time in milliseconds that the batch is allowed to run for.\n  Defaults to 1,000 (1 second). The default value is `1000`.","ref":"Pacer.Workflow.html#batch/3-batch-options"},{"type":"function","title":"Pacer.Workflow.execute/1","doc":"Takes a struct that has been defined via the `Pacer.Workflow.graph/1` macro.\n`execute` will run/execute all of the resolvers defined in the definition of the\n`graph` macro in an order that ensures all dependencies have been met before\nthe resolver runs.\n\nResolvers that have been defined within batches will be executed in parallel.","ref":"Pacer.Workflow.html#execute/1"},{"type":"macro","title":"Pacer.Workflow.field/2","doc":"The field/2 macro maps fields one-to-one to keys on the struct created via the graph definition.\n\nFields must be unique within a graph instance.","ref":"Pacer.Workflow.html#field/2"},{"type":"macro","title":"Options: - Pacer.Workflow.field/2","doc":"There are specific options that are allowed to be passed in to the field macro, as indicated below:\n\n* `:dependencies` (list of `t:atom/0`) - A list of dependencies from the graph.\n  Dependencies are specified as atoms, and each dependency must\n  be another field in the same graph.\n\n  Remember that cyclical dependencies are strictly not allowed,\n  so fields cannot declare dependencies on themselves nor on\n  any other field that has already declared a dependency on the\n  current field.\n\n  If the `dependencies` option is not given, it defaults to an\n  empty list, indicating that the field has no dependencies.\n  This will be the case if the field is a constant or can be\n  constructed from values already available in the environment.\n\n  The default value is `[]`.\n\n* `:doc` (`t:String.t/0`) - Allows users to document the field and provide background and/or context\n  on what the field is intended to be used for, what kind of data the field\n  contains, and how the data for the field is constructed.\n\n* `:resolver` (function of arity 1) - A resolver is a 1-arity function that specifies how to calculate\n  the value for a field.\n\n  The argument passed to the function will be a map that contains\n  all of the field's declared dependencies.\n\n  For example, if we have a field like this:\n\n  ```elixir\n  field(:request, resolver: &RequestHandler.resolve/1, dependencies: [:api_key, :url])\n  ```\n\n  The resolver `RequestHandler.resolve/1` would be passed a map that looks like this:\n  ```elixir\n  %{api_key: \" \", url: \"https://some.endpoint.com\"}\n  ```\n\n  If the field has no dependencies, the resolver will receive an empty map. Note though\n  that resolvers are only required for fields with no dependencies if the field is inside\n  of a batch. If your field has no dependencies and is not inside a batch, you can skip\n  defining a resolver and initialize your graph struct with a value that is either constant\n  or readily available where you are constructing the struct.\n\n  The result of the resolver will be placed on the graph struct under the field's key.\n\n  For the above, assuming a graph that looks like this:\n\n  ```elixir\n  defmodule MyGraph do\n    use Pacer.Workflow\n\n    graph do\n      field(:api_key)\n      field(:url)\n      field(:request, resolver: &RequestHandler.resolve/1, dependencies: [:api_key, :url])\n    end\n  end\n  ```\n\n  Then when the `RequestHandler.resolve/1` runs an returns a value of, let's say, `%{response: \"important response\"}`,\n  your graph struct would look like this:\n\n  ```elixir\n  %MyGraph{\n    api_key: \" \",\n    url: \"https://some.endpoint.com\",\n    request: %{response: \"important response\"}\n  }\n  ```\n\n* `:default` (`t:term/0`) - The default value for the field. If no default is given, the default value becomes\n  `#Pacer.Workflow.FieldNotSet `.\n\n* `:virtual?` (`t:boolean/0`) - A virtual field is used for intermediate or transient computation steps during the workflow and becomes a\n  node in the workflow's graph, but does not get returned in the results of the workflow execution.\n\n  In other words, virtual keys will not be included in the map returned by calling `Pacer.Workflow.execute/1`.\n\n  The intent of a virtual field is to allow a spot for intermediate and/or transient calculation steps but\n  to avoid the extra memory overhead that would be associated with carrying these values downstream if, for example,\n  the map returned from `Pacer.Workflow.execute/1` is stored in a long-lived process state; intermediate or transient\n  values can cause unnecessary memory bloat if they are carried into process state where they are not neeeded.\n\n  The default value is `false`.","ref":"Pacer.Workflow.html#field/2-options"},{"type":"function","title":"Pacer.Workflow.find_cycles/1","doc":"A Depth-First Search to find where is the dependency graph cycle\nand then display the cyclic dependencies back to the developer.","ref":"Pacer.Workflow.html#find_cycles/1"},{"type":"macro","title":"Pacer.Workflow.graph/1","doc":"The graph/1 macro is the main entrypoint into Pacer.Workflow to create a dependency graph struct.\n`use` the `Pacer.Workflow` macro at the top of your module and proceed to define your fields and/or batches.\n\nExample\n```elixir\n  defmodule MyValidGraph do\n    use Pacer.Workflow\n\n    graph do\n      field(:custom_field)\n      field(:field_a, resolver: &__MODULE__.do_work/1, dependencies: [:custom_field])\n      field(:field_with_default, default: \"this is a default value\")\n\n      batch :http_requests do\n        field(:request_1, resolver: &__MODULE__.do_work/1, dependencies: [:custom_field, :field_a])\n        field(:request_2, resolver: &__MODULE__.do_work/1)\n      end\n    end\n\n    def do_work(_), do: :ok\n  end\n```\n\nYour module may only define ONE graph per module.\n\nThe above example will also create a struct with all of the fields defined within the graph, as follows:\n\n```elixir\n%MyValidGraph{\n  custom_field: nil,\n  field_a: nil,\n  field_with_default: \"this is a default value\",\n  request_1: nil,\n  request_2: nil\n}\n```\n\n\nThe graph macro gives you access to some defined metadata functions, such as (using the above example graph):\n  - `MyValidGraph.__graph__(:fields)`\n  - `MyValidGraph.__graph__(:dependencies, :http_requests)`\n  - `MyValidGraph.__graph__(:resolver, :field_a)`\n\n**Caution: These metadata functions are mostly intended for Pacer's internal use. Do not rely on their return values in runtime\ncode as they may change as changes are made to the interface for Pacer.","ref":"Pacer.Workflow.html#graph/1"},{"type":"function","title":"Pacer.Workflow.validate_dependencies/1","doc":"","ref":"Pacer.Workflow.html#validate_dependencies/1"},{"type":"function","title":"Pacer.Workflow.validate_options/2","doc":"","ref":"Pacer.Workflow.html#validate_options/2"},{"type":"module","title":"Pacer.Workflow.FieldNotSet","doc":"Struct set as the initial value for fields in\na `Pacer.Workflow` definition when no explicit default\nvalue is provided.","ref":"Pacer.Workflow.FieldNotSet.html"},{"type":"module","title":"Pacer.Workflow.Testing","doc":"This module provides testing helpers for dependencies and fields for Pacer.Workflow.","ref":"Pacer.Workflow.Testing.html"},{"type":"module","title":"How to use in tests - Pacer.Workflow.Testing","doc":"`use` this module in tests:\n\n  use Pacer.Workflow.Testing\n\nThis will allow you to use the helper assertions in your test module.\n\nExamples:\n```elixir\ndefmodule SomeWorkflowTest do\n  use Pacer.Workflow.Testing\n\n  defmodule TestGraph do\n    use Pacer.Workflow\n\n    graph do\n      field(:field_a)\n\n      field(:field_with_default,\n        virtual?: true,\n        default: \"this is a default value\"\n      )\n\n      batch :http_requests do\n        field(:request_1,\n          resolver: &__MODULE__.do_work/1,\n          dependencies: [:field_a],\n          default: 2\n        )\n\n        field(:request_2,\n          resolver: &__MODULE__.do_work/1,\n          dependencies: [:field_a],\n          default: \"this is a default value for request2\"\n        )\n      end\n    end\n\n    def do_work(_), do: :ok\n  end\n\n  describe \"testing graph dependencies and fields\" do\n    test \"assert dependency of :field_a has no dependencies\" do\n      assert_dependencies(TestGraph, :field_a, [])\n    end\n\n    test \"assert batch dependency of :request_2\" do\n      assert_batch_dependencies(TestGraph, :request_2, [:field_a])\n    end\n\n    test \"provided fields match fields from workflow\" do\n      assert_fields(TestGraph, [\n        :field_a,\n        :field_with_default,\n        :request_1,\n        :request_2\n      ])\n    end\n  end\nend\n```","ref":"Pacer.Workflow.Testing.html#module-how-to-use-in-tests"},{"type":"function","title":"Pacer.Workflow.Testing.assert_batch_dependencies/3","doc":"Test helper to assert on batch field dependencies.\nFirst argument is your workflow, second is the batch field as an atom that you want to test,\nthird are the dependencies you are passing in to be dependencies for the field passed in as the second argument.\nNote, the third argument should be a list of atoms or an empty list in the case of no dependencies.","ref":"Pacer.Workflow.Testing.html#assert_batch_dependencies/3"},{"type":"function","title":"Example - Pacer.Workflow.Testing.assert_batch_dependencies/3","doc":"If we have a workflow defined as:\n```elixir\ndefmodule TestGraph do\n  use Pacer.Workflow\n\n  graph do\n    field(:field_a)\n\n    batch :http_requests do\n      field(:request_1,\n        resolver: &__MODULE__.do_work/1,\n        dependencies: [:field_a],\n        default: 2\n      )\n\n      field(:request_2,\n        resolver: &__MODULE__.do_work/1,\n        dependencies: [:field_a],\n        default: \"this is a default value for request2\"\n      )\n    end\n  end\n\n  def do_work(_), do: :ok\nend\n```\n\nWe can test the dependencies of both fields:\n\n```elixir\ntest \"assert batch fields have dependencies on :field_a\" do\n  assert_batch_dependencies(TestGraph, :request_1, [:field_a])\n  assert_batch_dependencies(TestGraph, :request_2, [:field_a])\nend\n```","ref":"Pacer.Workflow.Testing.html#assert_batch_dependencies/3-example"},{"type":"function","title":"Pacer.Workflow.Testing.assert_dependencies/3","doc":"Test helper to assert on non-batch field dependencies.\nFirst argument is your workflow, second is the field as an atom that you want to test,\nthird are the dependencies you are passing in to be dependencies for the field passed in as the second argument.\nNote, the third argument should be a list of atoms or an empty list in the case of no dependencies.","ref":"Pacer.Workflow.Testing.html#assert_dependencies/3"},{"type":"function","title":"Example - Pacer.Workflow.Testing.assert_dependencies/3","doc":"If we have a workflow defined as:\n```elixir\ndefmodule TestGraph do\n  use Pacer.Workflow\n\n  graph do\n    field(:field_a)\n\n    field(:field_b,\n      resolver: &__MODULE__.do_work/1,\n      dependencies: [:field_a]\n    )\n  end\nend\n```\n\nWe can test the dependencies of both fields:\n\n```elixir\ntest \"assert :field_a has no dependencies\" do\n  assert_dependencies(TestGraph, :field_a, [])\nend\n\ntest \"assert dependencies of :field_b\" do\n  assert_dependencies(TestGraph, :field_b, [:field_a])\nend\n```","ref":"Pacer.Workflow.Testing.html#assert_dependencies/3-example"},{"type":"function","title":"Pacer.Workflow.Testing.assert_fields/2","doc":"Test helper to assert on the field defined in the workflow.\nFirst argument is your workflow, second is a list of atoms that are fields in the workflow.\nOrdering does not matter in the list of atoms since we use MapSets to do the comparison.","ref":"Pacer.Workflow.Testing.html#assert_fields/2"},{"type":"function","title":"Example - Pacer.Workflow.Testing.assert_fields/2","doc":"If we have a workflow defined as:\n```elixir\ndefmodule TestGraph do\n  use Pacer.Workflow\n\n  graph do\n    field(:field_a)\n\n    field(:field_b,\n      resolver: &__MODULE__.do_work/1,\n      dependencies: [:field_a]\n    )\n\n    field(:field_with_default,\n      virtual?: true,\n      default: \"this is a default value\"\n    )\n\n    batch :http_requests do\n      field(:request_1,\n        resolver: &__MODULE__.do_work/1,\n        dependencies: [:field_a],\n        default: 2\n      )\n\n      field(:request_2,\n        resolver: &__MODULE__.do_work/1,\n        dependencies: [:field_a, :field_b],\n        default: \"this is a default value for request2\"\n      )\n    end\n  end\n\n  def do_work(_), do: :ok\nend\n```\n\nWe can test the fields:\n```elixir\ntest \"fields in the workflow\" do\n  assert_fields(TestGraph, [:field_a, :field_b, :request_2, :request_1])\nend\n```","ref":"Pacer.Workflow.Testing.html#assert_fields/2-example"},{"type":"module","title":"Pacer.Workflow.Visualization","doc":"A small wrapper around a Pacer Workflow's `visualization` metadata function.\n\nThis stringed strict diagraph that is returned can be used in graphviz or some other graph visualization tool.","ref":"Pacer.Workflow.Visualization.html"},{"type":"function","title":"Pacer.Workflow.Visualization.visualize/2","doc":"","ref":"Pacer.Workflow.Visualization.html#visualize/2"},{"type":"extras","title":"README","doc":"# Pacer: An Elixir Library for Dependency Graph-Based Workflows With Robust Compile Time Safety & Guarantees","ref":"readme.html"},{"type":"extras","title":"Overview - README","doc":"`Pacer.Workflow` is an abstraction designed for complex workflows where many interdependent data points need to be\n  stitched together to provide a final result, specifically workflows where each data point needs to be\n  loaded and/or calculated using discrete, application-specific logic.\n\n  See the moduledocs for `Pacer.Workflow` for detailed information about how to use Pacer and define your own\n  workflows, along with a detailed list of options and ideas that underlie Pacer.","ref":"readme.html#overview"},{"type":"extras","title":"Installation - README","doc":"Add :pacer to the list of dependencies in `mix.exs`:\n\n  ```elixir\n  [\n    {:pacer, \"~> 0.1\"}\n  ]\n  ```","ref":"readme.html#installation"},{"type":"extras","title":"Contributing - README","doc":"We welcome everyone to contribute to Pacer -- whether it is documentation updates, proposing and/or implementing new features, or contributing bugfixes. \n\n  Please feel free to create issues on the repo if you notice any bugs or if you would like to propose new features or implementations. \n\n  When contributing to the codebase, please:\n\n  1. Run the test suite locally with `mix test`\n  2. Verify Dialyzer still passes with `mix dialyzer`\n  3. Run `mix credo --strict`\n  4. Make sure your code has been formatted with `mix format`\n\n  In your PRs please provide the following detailed information as you see fit, especially for larger proposed changes:\n\n  1. What does your PR aim to do?\n  2. The reason/why for the changes\n  3. Validation and verification instructions (how can we verify that your changes are working as expected; if possible please provide one or two code samples that demonstrate the behavior)\n  4. Additional commentary if necessary -- tradeoffs, background context, etc.\n\n  We will try to provide responses and feedback in a timely manner. Please feel free to ping us if you have a PR or issue that has not been responded to.","ref":"readme.html#contributing"},{"type":"extras","title":"LICENSE","doc":"Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      \"License\" shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      \"Licensor\" shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      \"Legal Entity\" shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      \"control\" means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      \"You\" (or \"Your\") shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      \"Source\" form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      \"Object\" form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      \"Work\" shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      \"Derivative Works\" shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      \"Contribution\" shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, \"submitted\"\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as \"Not a Contribution.\"\n\n      \"Contributor\" shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a \"NOTICE\" text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an \"AS IS\" BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS\n\n  (C) Copyright 2023 Cars.com Inc. all rights reserved. \n\n   Licensed under the Apache License, Version 2.0 (the \"License\");\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an \"AS IS\" BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.","ref":"license.html"}],"content_type":"text/markdown"}